from laddu.amplitudes import breit_wigner, common, kmatrix, piecewise, ylm, zlm
from laddu.laddu import (
    Amplitude,
    AmplitudeID,
    Evaluator,
    Expression,
    Manager,
    Model,
    ParameterLike,
    constant,
    parameter,
)

__all__ = [
    'Amplitude',
    'AmplitudeID',
    'Evaluator',
    'Expression',
    'Manager',
    'Model',
    'ParameterLike',
    'breit_wigner',
    'common',
    'constant',
    'kmatrix',
    'parameter',
    'piecewise',
    'ylm',
    'zlm',
]



from typing import Literal

from laddu.amplitudes import Amplitude, ParameterLike
from laddu.utils.variables import Mass

def BreitWigner(
    name: str,
    mass: ParameterLike,
    width: ParameterLike,
    l: Literal[0, 1, 2, 3, 4],
    daughter_1_mass: Mass,
    daughter_2_mass: Mass,
    resonance_mass: Mass,
) -> Amplitude: ...



from typing import Literal, overload

from laddu.amplitudes import Amplitude
from laddu.utils.variables import Angles, Polarization

@overload
def Zlm(
    name: str,
    l: Literal[0],
    m: Literal[0],
    r: Literal['+', 'plus', 'pos', 'positive', '-', 'minus', 'neg', 'negative'],
    angles: Angles,
    polarization: Polarization,
) -> Amplitude: ...
@overload
def Zlm(
    name: str,
    l: Literal[1],
    m: Literal[-1, 0, 1],
    r: Literal['+', 'plus', 'pos', 'positive', '-', 'minus', 'neg', 'negative'],
    angles: Angles,
    polarization: Polarization,
) -> Amplitude: ...
@overload
def Zlm(
    name: str,
    l: Literal[2],
    m: Literal[-2, -1, 0, 1, 2],
    r: Literal['+', 'plus', 'pos', 'positive', '-', 'minus', 'neg', 'negative'],
    angles: Angles,
    polarization: Polarization,
) -> Amplitude: ...
@overload
def Zlm(
    name: str,
    l: Literal[3],
    m: Literal[-3, -2, -1, 0, 1, 2, 3],
    r: Literal['+', 'plus', 'pos', 'positive', '-', 'minus', 'neg', 'negative'],
    angles: Angles,
    polarization: Polarization,
) -> Amplitude: ...
@overload
def Zlm(
    name: str,
    l: Literal[4],
    m: Literal[-4, -3, -2, -1, 0, 1, 2, 3, 4],
    r: Literal['+', 'plus', 'pos', 'positive', '-', 'minus', 'neg', 'negative'],
    angles: Angles,
    polarization: Polarization,
) -> Amplitude: ...



from abc import ABCMeta, abstractmethod
from collections.abc import Sequence
from typing import Any, Literal

import numpy as np
import numpy.typing as npt

from laddu.amplitudes import Evaluator, Model
from laddu.data import Dataset

class Observer(metaclass=ABCMeta):
    @abstractmethod
    def callback(self, step: int, status: Status) -> tuple[Status, bool]: ...

class MCMCObserver(metaclass=ABCMeta):
    @abstractmethod
    def callback(self, step: int, ensemble: Ensemble) -> tuple[Ensemble, bool]: ...

class LikelihoodID:
    def __add__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __radd__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __mul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...
    def __rmul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...

class LikelihoodExpression:
    def __add__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __radd__(
        self, other: LikelihoodID | LikelihoodExpression | int
    ) -> LikelihoodExpression: ...
    def __mul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...
    def __rmul__(
        self, other: LikelihoodID | LikelihoodExpression
    ) -> LikelihoodExpression: ...

class LikelihoodTerm: ...

class LikelihoodManager:
    def __init__(self) -> None: ...
    def register(self, likelihood_term: LikelihoodTerm) -> LikelihoodID: ...
    def load(
        self, likelihood_expression: LikelihoodExpression
    ) -> LikelihoodEvaluator: ...

class LikelihoodEvaluator:
    parameters: list[str]
    def evaluate(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        threads: int | None = None,
    ) -> float: ...
    def evaluate_gradient(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def minimize(
        self,
        p0: list[float] | npt.NDArray[np.float64],
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['lbfgsb', 'nelder_mead'] = 'lbfgsb',
        max_steps: int = 4000,
        debug: bool = False,
        verbose: bool = False,
        **kwargs,  # noqa: ANN003
    ) -> Status: ...
    def mcmc(
        self,
        p0: list[list[float]] | npt.NDArray[np.float64],
        n_steps: int,
        *,
        method: Literal['ESS', 'AIES'] = 'ESS',
        debug: bool = False,
        verbose: bool = False,
        seed: int = 0,
        **kwargs,  # noqa: ANN003
    ) -> Ensemble: ...

class NLL:
    parameters: list[str]
    data: Dataset
    accmc: Dataset
    def __init__(
        self,
        model: Model,
        ds_data: Dataset,
        ds_accmc: Dataset,
    ) -> None: ...
    def as_term(self) -> LikelihoodTerm: ...
    def activate(self, name: str | list[str]) -> None: ...
    def activate_all(self) -> None: ...
    def deactivate(self, name: str | list[str]) -> None: ...
    def deactivate_all(self) -> None: ...
    def isolate(self, name: str | list[str]) -> None: ...
    def evaluate(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        threads: int | None = None,
    ) -> float: ...
    def evaluate_gradient(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def project(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        *,
        mc_evaluator: Evaluator | None = None,
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def project_with(
        self,
        parameters: list[float] | npt.NDArray[np.float64],
        name: str | list[str],
        *,
        mc_evaluator: Evaluator | None = None,
        threads: int | None = None,
    ) -> npt.NDArray[np.float64]: ...
    def minimize(
        self,
        p0: list[float] | npt.NDArray[np.float64],
        *,
        bounds: Sequence[tuple[float | None, float | None]] | None = None,
        method: Literal['lbfgsb', 'nelder_mead'] = 'lbfgsb',
        max_steps: int = 4000,
        debug: bool = False,
        verbose: bool = False,
        **kwargs,  # noqa: ANN003
    ) -> Status: ...
    def mcmc(
        self,
        p0: list[list[float]] | npt.NDArray[np.float64],
        n_steps: int,
        *,
        method: Literal['ESS', 'AIES'] = 'ESS',
        debug: bool = False,
        verbose: bool = False,
        seed: int = 0,
        **kwargs,  # noqa: ANN003
    ) -> Ensemble: ...

def LikelihoodScalar(name: str) -> LikelihoodTerm: ...

class Status:
    x: npt.NDArray[np.float64]
    err: npt.NDArray[np.float64] | None
    x0: npt.NDArray[np.float64]
    fx: float
    cov: npt.NDArray[np.float64] | None
    hess: npt.NDArray[np.float64] | None
    message: str
    converged: bool
    bounds: list[Bound] | None
    n_f_evals: int
    n_g_evals: int

    def __init__(self) -> None: ...
    def save_as(self, path: str) -> None: ...
    @staticmethod
    def load_from(path: str) -> Status: ...
    def as_dict(self) -> dict[str, Any]: ...
    def __getstate__(self) -> object: ...
    def __setstate__(self, state: object) -> None: ...

class Ensemble:
    dimension: tuple[int, int, int]

    def __init__(self) -> None: ...
    def save_as(self, path: str) -> None: ...
    @staticmethod
    def load_from(path: str) -> Status: ...
    def __getstate__(self) -> object: ...
    def __setstate__(self, state: object) -> None: ...
    def get_chain(self, *, burn: int = 0, thin: int = 1) -> npt.NDArray[np.float64]: ...
    def get_flat_chain(
        self, *, burn: int = 0, thin: int = 1
    ) -> npt.NDArray[np.float64]: ...
    def get_integrated_autocorrelation_times(
        self, *, c: float = 7.0, burn: int = 0, thin: int = 1
    ) -> npt.NDArray[np.float64]: ...

class Bound:
    lower: float
    upper: float

def integrated_autocorrelation_times(
    x: npt.NDArray[np.float64], *, c: float = 7.0
) -> npt.NDArray[np.float64]: ...

class AutocorrelationObserver:
    taus: npt.NDArray[np.float64]
    def __init__(
        self,
        *,
        n_check: int = 50,
        n_taus_threshold: int = 50,
        dtau_threshold: float = 0.01,
        discard: float = 0.5,
        terminate: bool = True,
        c: float = 7.0,
        verbose: bool = False,
    ) -> None: ...

__all__ = [
    'NLL',
    'AutocorrelationObserver',
    'Bound',
    'Ensemble',
    'LikelihoodEvaluator',
    'LikelihoodExpression',
    'LikelihoodID',
    'LikelihoodManager',
    'LikelihoodScalar',
    'LikelihoodTerm',
    'Status',
    'integrated_autocorrelation_times',
]



"""
Usage:
    amptools-to-laddu <input_file> <output_file> [--tree <treename>] [--pol-in-beam | --pol-angle <angle> --pol-magnitude <magnitude>] [-n <num-entries>]

Options:
    --tree <treename>            The tree name in the ROOT file [default: kin].
    --pol-in-beam                Use the beam's momentum for polarization (eps).
    --pol-angle <angle>          The polarization angle in degrees (only used if --pol-in-beam is not used)
    --pol-magnitude <magnitude>  The polarization magnitude (only used if --pol-in-beam is not used)
    -n <num-entries>             Truncate the file to the first n entries for testing.
"""  # noqa: D205, D400

from __future__ import annotations

from pathlib import Path

import numpy as np
import numpy.typing as npt
import pandas as pd
import uproot
from docopt import docopt
from loguru import logger


def read_root_file(
    input_path: Path | str,
    tree_name: str = 'kin',
    *,
    pol_in_beam: bool = False,
    pol_angle_rad: float | None = None,
    pol_magnitude: float | None = None,
    num_entries: int | None = None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    input_path = Path(input_path)
    """Read ROOT file and extract data with optional polarization from the beam."""
    logger.info(f'Reading ROOT file: {input_path}')
    tfile = uproot.open(input_path)
    tree = tfile[tree_name]
    logger.info(f'Using tree: {tree_name}')
    # Read necessary branches
    E_beam = tree['E_Beam'].array(library='np', entry_stop=num_entries)  # pyright: ignore
    Px_beam = tree['Px_Beam'].array(library='np', entry_stop=num_entries)  # pyright: ignore
    Py_beam = tree['Py_Beam'].array(library='np', entry_stop=num_entries)  # pyright: ignore
    Pz_beam = tree['Pz_Beam'].array(library='np', entry_stop=num_entries)  # pyright: ignore
    weight = (
        tree['Weight'].array(library='np', entry_stop=num_entries)  # pyright: ignore
        if 'Weight' in tree
        else np.ones_like(E_beam)
    )

    # Final state particles
    E_final = np.array(
        list(tree['E_FinalState'].array(library='np', entry_stop=num_entries))  # pyright: ignore
    )
    Px_final = np.array(
        list(tree['Px_FinalState'].array(library='np', entry_stop=num_entries))  # pyright: ignore
    )
    Py_final = np.array(
        list(tree['Py_FinalState'].array(library='np', entry_stop=num_entries))  # pyright: ignore
    )
    Pz_final = np.array(
        list(tree['Pz_FinalState'].array(library='np', entry_stop=num_entries))  # pyright: ignore
    )

    # Handle beam four-vector: (nevents, 4)
    p4_beam = np.stack([Px_beam, Py_beam, Pz_beam, E_beam], axis=-1)

    # Handle final state four-vectors: (nevents, nparticles, 4)
    p4_final = np.stack([Px_final, Py_final, Pz_final, E_final], axis=-1)

    # Check if EPS branch exists and update eps if needed
    if 'EPS' in tree:
        logger.info('EPS branch found. Using it for eps values.')
        eps = tree['EPS'].array(library='np', entry_stop=num_entries)  # pyright: ignore
        eps = eps[:, np.newaxis, :]
    if 'eps' in tree:
        logger.info('eps branch found. Using it for eps values.')
        eps = tree['eps'].array(library='np', entry_stop=num_entries)  # pyright: ignore
        eps = eps[:, np.newaxis, :]
    elif pol_in_beam:
        logger.info("Using beam's momentum for polarization (eps).")
        eps = np.stack([Px_beam, Py_beam, Pz_beam], axis=-1)[:, np.newaxis]
        # Reset beam momentum
        p4_beam[:, 0] = 0  # Set Px to 0
        p4_beam[:, 1] = 0  # Set Py to 0
        p4_beam[:, 2] = E_beam  # Set Pz = E for beam
    elif pol_angle_rad is not None and pol_magnitude is not None:
        logger.info(
            f'Using input polarization angle ({pol_angle_rad}) and magnitude ({pol_magnitude}).'
        )
        eps_x = pol_magnitude * np.cos(pol_angle_rad) * np.ones_like(E_beam)
        eps_y = pol_magnitude * np.sin(pol_angle_rad) * np.ones_like(E_beam)
        eps_z = np.zeros_like(E_beam)
        eps = np.stack([eps_x, eps_y, eps_z], axis=-1)[:, np.newaxis]
    else:
        logger.info('Using default or provided eps values.')
        eps = np.zeros((len(E_beam), 1, 3), dtype=np.float32)  # Default to 0

    # Concatenate the beam and final state particles: (nevents, nparticles+1, 4)
    logger.info('Concatenating beam and final state particles.')
    p4s = np.concatenate([p4_beam[:, np.newaxis, :], p4_final], axis=1)

    return p4s.astype(np.float32), eps.astype(np.float32), weight


def save_as_parquet(
    p4s: npt.NDArray, eps: npt.NDArray, weight: npt.NDArray, output_path: Path | str
) -> None:
    """Save the processed data into Parquet format."""
    logger.info('Saving data to Parquet format.')

    # Flatten the p4s and eps into individual columns
    columns = {}
    n_particles = p4s.shape[1]
    for i in range(n_particles):
        columns[f'p4_{i}_Px'] = p4s[:, i, 0]
        columns[f'p4_{i}_Py'] = p4s[:, i, 1]
        columns[f'p4_{i}_Pz'] = p4s[:, i, 2]
        columns[f'p4_{i}_E'] = p4s[:, i, 3]

    n_eps = eps.shape[1]
    for i in range(n_eps):
        columns[f'eps_{i}_x'] = eps[:, i, 0]
        columns[f'eps_{i}_y'] = eps[:, i, 1]
        columns[f'eps_{i}_z'] = eps[:, i, 2]

    # Add weights
    columns['weight'] = weight

    # Create a DataFrame and save as Parquet
    data = pd.DataFrame(columns)
    output_path = str(output_path)
    data.to_parquet(output_path, index=False)
    logger.info(f'File saved: {output_path}')


def convert_from_amptools(
    input_path: Path,
    output_path: Path,
    tree_name: str = 'kin',
    *,
    pol_in_beam: bool = False,
    pol_angle: float | None = None,
    pol_magnitude: float | None = None,
    num_entries: int | None = None,
) -> None:
    p4s, eps, weight = read_root_file(
        input_path,
        tree_name,
        pol_in_beam=pol_in_beam,
        pol_angle_rad=pol_angle,
        pol_magnitude=pol_magnitude,
        num_entries=num_entries,
    )
    save_as_parquet(p4s, eps, weight, output_path)


def run() -> None:
    args = docopt(__doc__ if __doc__ else '')
    input_file = args['<input_file>']
    output_file = args['<output_file>']
    tree_name = args['--tree']
    pol_in_beam = args['--pol-in-beam']
    pol_angle = float(args['--pol-angle']) * np.pi / 180 if args['--pol-angle'] else None
    pol_magnitude = float(args['--pol-magnitude']) if args['--pol-magnitude'] else None
    num_entries = int(args['-n']) if args['-n'] else None

    convert_from_amptools(
        Path(input_file),
        Path(output_file),
        tree_name,
        pol_in_beam=pol_in_beam,
        pol_angle=pol_angle,
        pol_magnitude=pol_magnitude,
        num_entries=num_entries,
    )

    df_read = pd.read_parquet(output_file)
    print('Output Parquet File (head):')
    print(df_read.head())
    print(f'Total Entries: {len(df_read)}')



from __future__ import annotations

from abc import ABCMeta, abstractmethod

from laddu.laddu import (
    NLL,
    AutocorrelationObserver,
    Bound,
    Ensemble,
    LikelihoodEvaluator,
    LikelihoodExpression,
    LikelihoodID,
    LikelihoodManager,
    LikelihoodScalar,
    LikelihoodTerm,
    Status,
    integrated_autocorrelation_times,
)


class Observer(metaclass=ABCMeta):
    @abstractmethod
    def callback(self, step: int, status: Status) -> tuple[Status, bool]:
        pass


class MCMCObserver(metaclass=ABCMeta):
    @abstractmethod
    def callback(self, step: int, ensemble: Ensemble) -> tuple[Ensemble, bool]:
        pass


__all__ = [
    'NLL',
    'AutocorrelationObserver',
    'Bound',
    'Ensemble',
    'LikelihoodEvaluator',
    'LikelihoodExpression',
    'LikelihoodID',
    'LikelihoodManager',
    'LikelihoodScalar',
    'LikelihoodTerm',
    'MCMCObserver',
    'Observer',
    'Status',
    'integrated_autocorrelation_times',
]



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from laddu.convert import read_root_file
from laddu.laddu import BinnedDataset, Dataset, Event, open
from laddu.utils.vectors import Vector3, Vector4

if TYPE_CHECKING:
    from pathlib import Path


def open_amptools(
    path: str | Path,
    tree: str = 'kin',
    *,
    pol_in_beam: bool = False,
    pol_angle: float | None = None,
    pol_magnitude: float | None = None,
    num_entries: int | None = None,
) -> Dataset:
    pol_angle_rad = pol_angle * np.pi / 180 if pol_angle else None
    p4s_list, eps_list, weight_list = read_root_file(
        path,
        tree,
        pol_in_beam=pol_in_beam,
        pol_angle_rad=pol_angle_rad,
        pol_magnitude=pol_magnitude,
        num_entries=num_entries,
    )
    return Dataset(
        [
            Event(
                [Vector4.from_array(p4) for p4 in p4s],
                [Vector3.from_array(eps_vec) for eps_vec in eps],
                weight,
            )
            for p4s, eps, weight in zip(p4s_list, eps_list, weight_list)
        ]
    )


__all__ = ['BinnedDataset', 'Dataset', 'Event', 'open']
